# Stepup

A simple control-flow library for node.JS that makes parallel execution, serial execution, and error handling painless.

Additionally, stepup provides error coalescing, [asynchronous try catch](http://github.com/Crabdude/trycatch), and long stack traces.

## How to install

	npm install stepup
	
Or, simply copy or link the lib/stepup.js file into your `$HOME/.node_libraries` folder.

## How to use

The stepup library exports a single function I call `step`.	It accepts any number of functions as arguments and runs them in serial order using the passed in `this` context as the callback to the next step.

	step(
		function readSelf() {
		fs.readFile(__filename, this);
		},
		function capitalize(err, text) {
		if (err) throw err;
		return text.toUpperCase();
		},
		function showIt(err, newText) {
		if (err) throw err;
		console.log(newText);
		}
	);

Notice that we pass in `this` as the callback to `fs.readFile`.	When the file read completes, Stepup will send the result as the arguments to the next function in the chain.	Then in the `capitalize` function we're doing synchronous work so we can simple return the new value and Stepup will route it as if we called the callback.

The first parameter is reserved for errors since this is the node standard.	Also any exceptions thrown are caught and passed as the first argument to the next function.	As long as you don't nest callback functions inline your main functions this prevents there from ever being any uncaught exceptions.	This is very important for long running node.JS servers since a single uncaught exception can bring the whole server down.

Also there is support for parallel actions:

	step(
		// Loads two files in parallel
		function loadStuff() {
		fs.readFile(__filename, this.parallel());
		fs.readFile("/etc/passwd", this.parallel());
		},
		// Show the result when done
		function showStuff(err, code, users) {
		if (err) throw err;
		sys.puts(code);
		sys.puts(users);
		}
	)

Here we pass `this.parallel()` instead of `this` as the callback.	It internally keeps track of the number of callbacks issued and preserves their order then giving the result to the next step after all have finished.	If there is an error in any of the parallel actions, it will be passed as the first argument to the next step.

Also you can use group with a dynamic number of common tasks.

	step(
		function readDir() {
		fs.readdir(__dirname, this);
		},
		function readFiles(err, results) {
		if (err) throw err;
		// Create a new group
		var group = this.group();
		results.forEach(function (filename) {
			if (/\.js$/.test(filename)) {
			fs.readFile(__dirname + "/" + filename, 'utf8', group());
			}
		});
		},
		function showAll(err , files) {
		if (err) throw err;
		sys.p(files);
		}
	);

*Note* that we both call `this.group()` and `group()`.	The first reserves a slot in the parameters of the next step, then calling `group()` generates the individual callbacks and increments the internal counter.

<a name="pookie"/>
### Error Handling

Stepup also supports robust asynchronous error handling and error coalescing. If the first step function contains one or more arguments, it is treated as the error handler for al steps. Any error generated by a step will coalesce up to the error handler with the option to continue by calling `next` or `this`. When an errorHandler is provided, errors will not be passed into steps as they normally would.

	step(
		function errorHandler(err, next) {
		// log and ignore
		console.log('\nThis is the long stack trace from an async error: \n', err.stack);
		next('some data');
		},
		function readSelf() {
		fs.readFile(__filename, this);
		},
		function capitalize(text) {
		return (''+text).toUpperCase();
		},
		function showIt(newText) {
		console.log(newText);
		throw new Error('fail.');
		},
		function afterError(data) {
			console.log('\nThis is after errorHandler has been called: \n', data);
		}
	);

Output:

	stepup/examples$ node errorHandler.js 
	VAR FS = REQUIRE('FS');
	VAR STEP = REQUIRE('../LIB/STEPUP');
	
	STEP(
	  FUNCTION ERRORHANDLER(ERR, NEXT) {
	    // LOG AND IGNORE
	    CONSOLE.LOG('\NTHIS IS THE LONG STACK TRACE FROM AN ASYNC ERROR: \N', ERR.STACK);
	    NEXT('SOME DATA');
	  },
	  FUNCTION READSELF() {
	    FS.READFILE(__FILENAME, THIS);
	  },
	  FUNCTION CAPITALIZE(TEXT) {
	    RETURN (''+TEXT).TOUPPERCASE();
	  },
	  FUNCTION SHOWIT(NEWTEXT) {
	    CONSOLE.LOG(NEWTEXT);
	    THROW NEW ERROR('FAIL.');
	  },
	  FUNCTION AFTERERROR(DATA) {
	  	CONSOLE.LOG('\NTHIS IS AFTER ERRORHANDLER HAS BEEN CALLED: \N', DATA);
	  }
	);
	
	This is the long stack trace from an async error: 
	 Error: fail.
	    at Function.showIt (/Users/adamcrabtree/projects/mine/stepup/examples/errorHandler.js:18:11)
	    at run (/Users/adamcrabtree/projects/mine/stepup/lib/stepup.js:45:20)
	    at next (/Users/adamcrabtree/projects/mine/stepup/lib/stepup.js:55:4)
	    at run (/Users/adamcrabtree/projects/mine/stepup/lib/stepup.js:48:5)
	    at next (/Users/adamcrabtree/projects/mine/stepup/lib/stepup.js:55:4)
	    at [object Object].<anonymous> (fs.js:107:5)
	    ----------------------------------------
	    at fs.readFile
	    at Function.readSelf (/Users/adamcrabtree/projects/mine/stepup/examples/errorHandler.js:11:8)
	    at run (/Users/adamcrabtree/projects/mine/stepup/lib/stepup.js:45:20)
	    at next (/Users/adamcrabtree/projects/mine/stepup/lib/stepup.js:55:4)
	    at Step (/Users/adamcrabtree/projects/mine/stepup/lib/stepup.js:124:3)
	    at Object.<anonymous> (/Users/adamcrabtree/projects/mine/stepup/examples/errorHandler.js:4:1)
	
	This is after errorHandler has been called: 
	 some data